import "dart:io";
import "package:path/path.dart" as path;
import "package:ebisu/ebisu.dart";
import "package:ebisu/ebisu_dart_meta.dart";
import "package:logging/logging.dart";

String _topDir;

void main() {

  Logger.root.onRecord.listen((LogRecord r) =>
      print("${r.loggerName} [${r.level}]:\t${r.message}"));
  String here = path.absolute(Platform.script.path);
  _topDir = path.dirname(path.dirname(here));
  useDartFormatter = true;
  System ebisu = system('ebisu_cpp')
    ..includesHop = true
    ..license = 'boost'
    ..pubSpec.homepage = 'https://github.com/patefacio/ebisu_cpp'
    ..pubSpec.version = '0.0.23'
    ..pubSpec.doc = 'A library that supports code generation of cpp and others'
    ..pubSpec.addDependency(new PubDependency('path')..version = ">=1.3.0<1.4.0")
    ..pubSpec.addDevDependency(new PubDependency('unittest'))
    ..rootPath = _topDir
    ..doc = 'A library that supports code generation of cpp'
    ..testLibraries = [
      library('test_cpp_enum'),
      library('test_cpp_member'),
      library('test_cpp_class'),
      library('test_cpp_method'),
      library('test_cpp_utils'),
      library('test_cpp_namer'),
      library('test_cpp_generic'),
      library('test_cpp_test_provider'),
      library('test_hdf5_support'),
    ]
    ..libraries = [
      library('ebisu_cpp')
      ..doc = cppLibraryDoc
      ..includesLogger = true
      ..imports = [
        'package:id/id.dart',
        'package:ebisu/ebisu.dart',
        'package:quiver/iterables.dart',
        "'package:path/path.dart' as path",
        'io',
        'collection',
        "'dart:math' hide max",
      ]
      ..enums = [
        enum_('access')
        ..doc = '''
Access designation for a class member variable.

C++ supports *public*, *private* and *protected* designations. This designation
is at a higher abstraction in that selecting access determines both the c++
protection as well as the set of accessors that are generated. *Member*
instances have both *access* and *cppAccess* so full range is available, but
general cases should be covered by setting only the *access* variable of a
member.

* IA: *inaccessible* which provides no accessor with c++ access *private*.
  It is aliased to *ia*.

* RO: *read-only* which provides a read accessor with c++ access *private*.
  It is aliased to *ro*.

* RW: *read-write* which provides read and write accessor with c++ access *private*.
  It is aliased to *rw*.

* WO: *write-only* which provides write access and the c++ access is *private*.
  It is aliased to *wo*. It may sound counterintuitive, but a use case for this
  might be to accept the generated write accessor but hand code a read accessor
  requiring special logic.

Note: If the desire is to have a public member that is public and has no
accessors, set *cppAccess* to *public* andd set *access* to null.

# Examples

*cppAccess* null with *access* of *ro* gives:

    class C_1
    {
    public:
      //! getter for x_ (access is Ro)
      std::string const& x() const { return x_; }
    private:
      std::string x_ {};
    };

*cppAccess* CppAccess.protected with *access* of *ro* gives:

    class C_1
    {
    public:
      //! getter for x_ (access is Ro)
      std::string const& x() const { return x_; }
    protected:
      std::string x_ {};
    };
'''
        ..hasLibraryScopedValues = true
        ..values = [
          enumValue(id('ia'))
          ..doc = '**Inaccessible**. Designates a member that is *private* by default and no accessors',
          enumValue(id('ro'))
          ..doc = '**Read-Only**. Designates a member tht is *private* by default and a read accessor',
          enumValue(id('rw'))
          ..doc = '**Read-Write**. Designates a member tht is *private* by default and both read and write accessors',
          enumValue(id('wo'))
          ..doc = '''
**Write-Only**. Designates a member tht is *private* by default and
write accessor only.  Useful if you want the standard write accessor
but a custom reader.''',
        ],
        enum_('cpp_access')
        ..doc = '''
Cpp access designations:

  * public
  * private
  * protected

This designation is used in multiple contexts such as:

  * Overriding the protection of a member
  * On *Base* instances to indicate the access associated with inheritance
  * On class methods (ctor, dtor, ...) to designate access
'''
        ..hasLibraryScopedValues = true
        ..values = [
          enumValue(id('public'))
          ..doc = 'C++ public designation',
          enumValue(id('protected'))
          ..doc = 'C++ protected designation',
          enumValue(id('private'))
          ..doc = 'C++ private designation',
        ],
        enum_('ref_type')
        ..doc = 'Reference type'
        ..hasLibraryScopedValues = true
        ..values = [
          enumValue(id('ref'))
          ..doc = 'Indicates a reference to type: *T &*',
          enumValue(id('cref'))
          ..doc = 'Indicates a const reference to type: *T const&*',
          enumValue(id('vref'))
          ..doc = 'Indicates a volatile reference to type: *T volatile&*',
          enumValue(id('cvref'))
          ..doc = 'Indicates a const volatile reference to type: *T const volatile&*',
          enumValue(id('value'))
          ..doc = 'Indicates not a reference'
        ],
        enum_('ptr_type')
        ..doc = 'Standard pointer type declaration'
        ..hasLibraryScopedValues = true
        ..values = [
          enumValue(id('sptr'))
          ..doc = 'Indicates *std::shared_ptr< T >*',
          enumValue(id('uptr'))
          ..doc = 'Indicates *std::unique_ptr< T >*',
          enumValue(id('scptr'))
          ..doc = 'Indicates *std::shared_ptr< const T >*',
          enumValue(id('ucptr'))
          ..doc = 'Indicates *std::unique_ptr< const T >*',
        ],
        enum_('template_parm_type')
        ..values = [
          enumValue('type'),
          enumValue('non_type'),
        ],
      ]
      ..classes = [

        class_('installation_decorator')
        ..isAbstract = true
        ..doc = '''
Establishes an interface to allow decoration of classes and updates
(primarily additions) to an [Installation].
''',

        class_('entity')
        ..isAbstract = true
        ..doc = """
Exposes common elements for named entities, including their [id] and
documentation. Additionally tracks parentage/ownership of entities.

This is abstract for purposes of ownership. Each [Entity] knows its
owning entity up until [Installation] which is the root entity. A call
to [generate] on [Installation] will [setOwnership] which subclasses
can trick down establishing ownership.

The purpose of linking all [Entity] instances in a virtual tree type
structure is so lookups can be done for entities.

[Entity] must be created with an argument representing an Id.  That
argument may be a string, in which case it is converted to an [Id].
That argument may be an [Id].

For many/most [Entity] subclasses there is often a corresponding
method that simply creates in instance of the subclass. For example,

    class Lib extends Entity... {
       Lib(Id id) : super(id);
       ...
    }

    Lib lib(Object id) => new Lib(id is Id ? id : new Id(id));

This now allows this approach:

      final myLib = lib('my_awesome_lib')
        ..headers = [
          header('my_header')
          ..classes = [
            class_('my_class')
            ..members = [
              member('my_member')
            ]
          ]
        ];

      print(myLib);

prints:

    lib(myAwesomeLib)
      headers:
        header(myHeader)
          classes:[My_class]

      tests:
"""
        ..members = [
          member('id')
          ..doc = 'Id for the entity'
          ..type = 'Id',
          member('brief')
          ..doc = 'Brief description for the entity',
          member('descr')
          ..doc = 'Description of entity',
          member('owner')
          ..doc = '''
The entity containing this entity (e.g. the [Class] containing the [Member]).
[Installation] is a top level entity and has no owner.
'''
          ..access = RO
          ..type = 'Entity',
          member('entity_path')
          ..doc = 'Path from root to this entity'
          ..type = 'List<Entity>'
          ..access = RO
          ..classInit = [],
          member('namer')
          ..doc = '''
Namer to be used when generating names during generation. There is a
default namer, [EbisuCppNamer] that is used if one is not provide. To
create your own naming conventions, provide an implementation of
[Namer] and set an assign that namer to a top-level [Entity], such as
the [Installation]. The assigned namer will be propogated to all
genration utilities.
'''
          ..access = IA
          ..type = 'Namer',
        ],
        class_('using')
        ..doc = 'Object corresponding to a using statement'
        ..extend = 'Entity'
        ..members = [
          member('rhs')
          ..doc = '''
The right hand side of using (ie the type decl being named)'''
          ..access = RO,
        ],
        class_('template_parm')
        ..isAbstract = true
        ..extend = 'Entity',
        class_('type_template_parm')
        ..extend = 'TemplateParm'
        ..members = [
          member('type_id'),
        ],
        class_('decl_template_parm')
        ..extend = 'TemplateParm'
        ..members = [
          member('terms')..type = 'List<String>',
          member('id_index')..doc = 'Index into the terms indicating the id'
          ..type = 'int',
        ],
        class_('template')
        ..extend = 'Entity'
        ..doc = '''
Represents a template declaration comprized of a list of [decls]
'''
        ..members = [
          member('parms')
          ..type = 'List<TemplateParm>',
        ],
      ]
      ..parts = [
        part('generic')
        ..classes = [
          class_('traits')
          ..members = [
            member('usings')..type = 'Map<String, Using>'..classInit = {},
            member('const_exprs')..type = 'List<ConstExpr>'..classInit = [],
          ],
          class_('traits_requirements')
          ..doc = 'Collection of requirements for a [Traits] entry in a [TraitsFamily]'
          ..members = [
            member('usings')..type = 'List<Id>'..access = RO,
            member('const_exprs')..type = 'List<Id>'..access = RO,
          ],
          class_('traits_family')
          ..extend = 'Entity'
          ..members = [
            member('traits_requirements')..type = 'TraitsRequirements',
            member('traits')..type = 'List<Traits>',
          ],
        ],
        part('utils')
        ..classes = [

          class_('cpp_logger')
          ..doc = '''
Establishes an abstract interface to provide customizable c++ log messages

Not wanting to commit to a single logging solution, this class allows
client code to make certain items [Loggable] and not tie the generated
code to a particular logging solution. A default [CppLogger] that makes
use of *spdlog* is provided.
'''
          ..isAbstract = true,

          class_('loggable')
          ..doc = '''
Mixin to indicate an item is loggable.

Examples might be member accessors, member constructors, etc
'''
          ..members = [
            member('is_logged')
            ..doc = 'If true the [Loggable] item is logged'
            ..classInit = false,
          ],

          class_('const_expr')
          ..doc = """
Simple variable constexprs.

      print(new ConstExpr('secret', 42));
      print(new ConstExpr(new Id('voo_doo'), 'foo'));
      print(new ConstExpr('pi', 3.14));

prints:

    constexpr int Secret { 42 };
    constexpr char const* Voo_doo { "foo" };
    constexpr double Pi { 3.14 };
"""
          ..extend = 'Entity'
          ..members = [
            member('type')
            ..doc = 'The c++ type of the constexpr',
            member('value')
            ..doc = 'The initialization for the constexpr'
            ..type = 'Object'
            ..access = IA,
            member('namespace')
            ..doc = 'Any namespace to wrap the constexpr in'
            ..type = 'Namespace',
            member('is_class_scoped')
            ..doc = 'If class scoped the expr should be static'
            ..classInit = false,
            member('is_hex')
            ..doc = '''
If true and literal is numeric it is assigned as hex.
The idea is to make C++ more readable when large constants are used.
'''
            ..classInit = false,
          ],
          class_('forward_decl')
          ..doc = 'A forward declaration'
          ..hasCtorSansNew = true
          ..members = [
            member('type')
            ..doc = 'The c++ type being forward declared'
            ..ctors = [''],
            member('namespace')
            ..doc = 'The namespace to which the class being forward declared belongs'
            ..type = 'Namespace'..ctorsOpt = [''],
          ],
          class_('code_generator')
          ..doc = 'Establishes an interface for generating code'
          ..isAbstract = true,
          class_('friend_class_decl')
          ..doc = 'Friend class declaration'
          ..isImmutable = true
          ..hasCtorSansNew = true
          ..members = [
            member('decl')
            ..doc = 'Declaration text without the *friend* and *class* keywords'
          ],
          class_('namespace')
          ..doc = 'Represents a c++ namespace which is essentially a list of names'
          ..members = [
            member('names')
            ..doc = 'The individual names in the namespace'
            ..type = 'List<String>'..classInit = [],
          ],
          class_('includes')
          ..doc = 'Collection of header includes'
          ..members = [
            member('included')
            ..doc = 'Set of strings representing the includes'
            ..access = RO
            ..type = 'Set<String>'
          ],
          class_('namer')
          ..doc = 'Provides support for consistent naming of C++ entities'
          ..isAbstract = true,
          class_('ebisu_cpp_namer')
          ..implement = [ 'Namer' ]
          ..doc = '''
Default namer establishing reasonable conventions, that are fairly
*snake* case heavy like the STL.
''',
          class_('google_namer')
          ..implement = [ 'Namer' ]
          ..doc = 'Namer based on google coding conventions',
          class_('base')
          ..doc = '''
A base class of another class.


The style of inheritance is determined by [virtual] and [access]. Examples:

Default is *not* virtual and [public] inheritance:

    class_('derived')
    ..bases = [
      base('Base')
    ];

gives:

    class Derived : public Base {};

With overrides:

    class_('derived')
    ..bases = [
      base('Base')
      ..isVirtual = true
      ..access = protected
    ];

Gives:

    class Derived :
      protected virtual Base
    {
    };
'''
          ..hasCtorSansNew = true
          ..members = [
            member('class_name')
            ..doc = 'The name of the class being derived from'
            ..ctors = [''],
            member('access')
            ..doc = 'Is base class public, protected, or private'
            ..type = 'CppAccess'..classInit = 'public',
            member('init')
            ..doc = 'How to initiailize the base class in ctor initializer',
            member('is_virtual')
            ..doc = 'If true inheritance is virtual'
            ..classInit = false,
            member('is_streamable')
            ..doc = 'If true and streamers are being provided, base is streamed first'
            ..classInit = false,
          ]
        ],
        part('file')
        ..classes = [
          class_('cpp_file')
          ..doc = '''
Establishes an interface and common elements for c++ file, such as
*Header* and *Impl*.'''
          ..isAbstract = true
          ..extend = 'Entity'
          ..mixins = [ 'Testable' ]
          ..members = [
            member('namespace')
            ..doc = 'Namespace associated with this file'
            ..type = 'Namespace',
            member('custom_blocks')
            ..doc = '''
List of blocks requiring custom code and therefore inserted into the
file with *Protect Blocks*. Note it is a list of *FileCodeBlock*
enumeration values. *CodeBlocks* can be used to inject code into
the location designated by their value. Additionally *CodeBlocks*
have support for a single custom *Protect Block*'''
            ..type = 'List<FileCodeBlock>'..classInit = [],
            member('code_blocks')
            ..doc = '''
Mapping of the *FileCodeBlock* to the corresponding *CodeBlock*.'''
            ..type = 'Map<FileCodeBlock, CodeBlock>'..access = IA..classInit = {},
            member('classes')
            ..doc = 'List of classes whose definitions are included in this file'
            ..type = 'List<Class>'..classInit = [],
            member('includes')
            ..doc = 'List of includes required by this c++ file'
            ..type = 'Includes'..access = RO..classInit = 'new Includes()',
            member('const_exprs')
            ..doc = 'List of c++ *constexprs* that will appear near the top of the file'
            ..type = 'List<ConstExpr>'..classInit = [],
            member('forward_decls')
            ..doc = 'List of forward declarations that will appear near the top of the file'
            ..type = 'List<ForwardDecl>'..classInit = [],
            member('usings')
            ..doc = 'List of using statements that will appear near the top of the file'
            ..type = 'List<Using>'..access = RO..classInit = [],
            member('enums')
            ..doc = 'List of enumerations that will appear near the top of the file'
            ..type = 'List<Enum>'..classInit = [],
            member('interfaces')
            ..doc = '''
List of interfaces for this header. Interfaces result in either:

* abstract base class with pure virtual methods
* static polymorphic base class with inline forwarding methods
'''
            ..type = 'List<Interface>'..classInit = [],
            member('basename')
            ..access = RO,
            member('file_path')..access = RO,
          ],
        ],
        part('enum')
        ..classes = [
          class_('enum')
          ..doc = """
A c++ enumeration.

There are two main styles of enumerations, *standard* and
*mask*.

Enumerations are often used to establish values to be used in
masks. The actual manipulation with masks is done in the *int* space,
but enums are convenient for setting up the mask values:

      print(enum_('gl_buffer')
          ..values = [ 'gl_color_buffer', 'gl_depth_buffer',
            'gl_accum_buffer', 'gl_stencil_buffer' ]
          ..isMask = true);

will print:

    enum Gl_buffer {
      Gl_color_buffer_e = 1 << 0,
      Gl_depth_buffer_e = 1 << 1,
      Gl_accum_buffer_e = 1 << 2,
      Gl_stencil_buffer_e = 1 << 3
    };

The *values* for enumeration entries can be ignored when their only
purpose is to draw distinction:

    print(enum_('region')..values = ['north', 'south', 'east', 'west']);

will print:

    enum Region {
      North_e,
      South_e,
      East_e,
      West_e
    };

Sometimes it is important not only to distinguish, but also to assign
values. For this purpose the values associated with the entries may be
provided via the [valueMap]

    print(enum_('thresholds')
          ..valueMap = { 'high' : 100, 'medium' : 50, 'low' : 10 });

gives:

    enum Thresholds {
      High_e = 100,
      Medium_e = 50,
      Low_e = 10
    };

Optionally the [isClass] field can be set to improve scoping by making
the enum a *class* enum.

    print(enum_('color_as_class')
          ..values = ['red', 'green', 'blue']
          ..isClass = true);

gives:

    enum class Color_as_class {
      Red_e,
      Green_e,
      Blue_e
    };

Optionally the [enumBase] can be used to specify the
base type. This is particularly useful where the enum
is a field in a *packed* structure.

    print(enum_('color_with_base')
          ..values = ['red', 'green', 'blue']
          ..enumBase = 'std::int8_t');

gives:

    enum Color_with_base : std::int8_t {
      Red_e,
      Green_e,
      Blue_e
    };

[isClass] may be combined with [enumBase].

The [isStreamable] flag will provide *to_c_str* and *operator<<* methods:

    print(enum_('color')
          ..values = ['red', 'green', 'blue']
          ..enumBase = 'std::int8_t'
          ..isStreamable = true
          );

gives:

    enum class Color : std::int8_t {
      Red_e,
      Green_e,
      Blue_e
    };
    inline char const* to_c_str(Color e) {
      switch(e) {
        case Color::Red_e: return "Red_e";
        case Color::Green_e: return "Green_e";
        case Color::Blue_e: return "Blue_e";
      }
    }
    inline std::ostream& operator<<(std::ostream &out, Color e) {
      return out << to_c_str(e);
    }

For the *standard* style enum you can use the [hasFromCStr] to include
a c-string to enum conversion method:

    inline void from_c_str(char const* str, Color &e) {
      using namespace std;
      if(0 == strcmp("Red_e", str)) { e = Color::Red_e; return; }
      if(0 == strcmp("Green_e", str)) { e = Color::Green_e; return; }
      if(0 == strcmp("Blue_e", str)) { e = Color::Blue_e; return; }
      string msg { "No Color matching:" };
      throw std::runtime_error(msg + str);
    }
"""
          ..extend = 'Entity'
          ..members = [
            member('values')
            ..doc = 'Strings for the values of the enum'
            ..type = 'List<String>'
            ..access = RO,
            member('ids')
            ..doc = 'Ids for the values of the enum'
            ..type = 'List<Id>'
            ..access = IA,
            member('value_names')
            ..doc = 'Names for values as they appear'
            ..type = 'List<String>'
            ..access = IA,
            member('value_map')
            ..doc = 'String value, numeric value pairs'
            ..type = 'Map<String, int>'
            ..access = RO,
            member('is_class')
            ..doc = 'If true the enum is a class enum as opposed to "plain" enum'
            ..classInit = false,
            member('enum_base')
            ..doc = 'Base of enum - if set must be an integral type',
            member('has_from_c_str')
            ..doc = 'If true adds from_c_str method'..classInit = false,
            member('has_to_c_str')
            ..doc = 'If true adds to_c_str method'..classInit = false,
            member('is_streamable')
            ..doc = 'If true adds streaming support'..classInit = false,
            member('is_mask')
            ..doc = 'If true the values are powers of two for bit masking'
            ..classInit = false,
            member('is_nested')
            ..doc = 'If true is nested in class and requires *friend* stream support'
            ..classInit = false,
            member('is_displayed_hex')
            ..doc = '''
If the map has values assigned by user, this can be used to display
them in the enum as hex'''
            ..classInit = false,
          ],
        ],
        part('member')
        ..classes = [
          class_('member')
          ..doc = '''
A member or field included in a class.

## Basics

Members are typed (i.e. have [type]) and optionally initialized.

For example:

    member('widget')..type = 'Widget'..init = 'Widget()'

gives:

    Widget widget_ { Widget() };

For some C++ types (*double*, *int*, *std::string*, *bool*) the type
can be inferred if a suitable [init] is provided:

Examples:

    member('number')..init = 4
    member('pi')..init = 3.14
    member('default_tag')..init = 'empty'
    member('is_strong')..init = false

give respectively:

    int number_ { 4 };
    double pi_ { 3.14 };
    std::string default_tag_ { "empty" };
    bool is_strong_ { false };

## Encapsulation

Encapsulation can be achieved by setting [access] and/or
[cppAccess]. Setting [access] is the preferred approach since it
provides a consistent, sensible pattern for hiding and accessing
members (See [Access]).

*Read-Only* Example:

    (class_('c')
        ..members = [
          member('readable')..init = 'foo'..access = ro
        ])
    .definition

Gives:

    class C
    {
    public:
      //! getter for readable_ (access is Ro)
      std::string const& readable() const { return readable_; }
    private:
      std::string readable_ { "foo" };
    };


*Inaccessible* Example:

    (class_('c')
        ..members = [
          member('inaccessible')..init = 'foo'..access = ia
        ])
    .definition

Gives:

    class C
    {
    private:
      std::string inaccessible_ { "foo" };
    };

*Read-Write* Example:

    (class_('c')
      ..members = [
        member('read_write')..init = 'foo'..access = rw
      ])
    .definition

Gives:

    class C
    {
    public:
      //! getter for read_write_ (access is Rw)
      std::string const& read_write() const { return read_write_; }
      //! setter for read_write_ (access is Access.rw)
      void read_write(std::string & read_write) { read_write_ = read_write; }
    private:
      std::string read_write_ { "foo" };
    };


Note that read-write keeps the member *private* by default and allows
access through methods. However, complete control over C++ access of
members can be obtained with [cppAccess]. Here are two such examples:

No accessors, just C++ public:

    (class_('c')
      ..members = [
        member('full_control')..init = 'foo'..cppAccess = public
      ])
    .definition

Gives:

    class C
    {
    public:
      std::string full_control { "foo" };
    };

Finally, using both [access] and [cppAccess] for more control:

    (class_('c')
      ..members = [
        member('more_control')..init = 'foo'..access = ro..cppAccess = protected
      ])
    .definition


Gives:

    class C
    {
    public:
      //! getter for more_control_ (access is Ro)
      std::string const& more_control() const { return more_control_; }
    protected:
      std::string more_control_ { "foo" };
    };

'''
          ..extend = 'Entity'
          ..members = [
            member('type')..doc = 'Type of member',
            member('init')
            ..doc = """
Initialization of member.

If [type] of [Member] is null and [init] is set with a Dart type
which can reasonably map to a C++ type, then type is inferred.
Currently the mappings are:
    {
      int : int,
      double : double,
      string : std::string,
      bool : bool,
      List(...) : std::vector<...>,
    }

For example:

    member('name')..init = 'UNASSIGNED' => name is std::string
    member('x')..init = 0               => x is int
    member('pi')..init = 3.14           => pi is double

"""..access = RO,
            member('ctor_init')
            ..doc = '''
Rare usage - member b depends on member a (e.g. b is just a string rep of a
which is int), a is passed in for construction but b can be initialized directly
from a. If ctorInit is set on a member, any memberCtor will include this text to
initialize it''',
            member('access')
            ..doc = 'Idiomatic access of member'
            ..type = 'Access'
            ..access = WO,
            member('cpp_access')
            ..doc = 'C++ style access of member'..type = 'CppAccess'..access = WO,
            member('ref_type')
            ..doc = 'Ref type of member'..type = 'RefType'..classInit = 'value',
            member('is_by_ref')
            ..doc = 'Pass member around by reference'..type = 'bool'..access = WO,
            member('is_static')..doc = 'Is the member static'
            ..classInit = false,
            member('is_mutable')..doc = 'Is the member mutable'
            ..classInit = false,
            member('is_const')..doc = 'Is the member const'
            ..classInit = false
            ..access = WO,
            member('is_const_expr')..doc = 'Is the member a constexprt'
            ..classInit = false,
            member('has_no_init')
            ..doc = 'If set will not initialize variable - use sparingly'
            ..classInit = false,
            member('is_serialized_as_int')
            ..doc = 'Indicates this member is an enum and if serialized should be serialized as int'
            ..classInit = false,
            member('is_cereal_transient')
            ..doc = 'Indicates this member should not be serialized via cereal'
            ..classInit = false,
            member('getter_return_modifier')
            ..doc = '''
A function that may be used to modify the value returned from a
getter.  If a modifier function of type [GetReturnModifier] is
provided it will be used to update what the accessor returns.

For example:

    print(clangFormat(
            (member('message_length')
                ..type = 'int32_t'
                ..access = ro
                ..getterReturnModifier =
                  ((member, oldValue) => 'endian_convert(\$oldValue)'))
            .getter));

prints:

    //! getter for message_length_ (access is Ro)
    int32_t message_length() const { return endian_convert(message_length_); }

Notes: No required *parens* when used inline with cascades. A trailing
semicolon is *not* required and the modifier accessor must return the
same type as the member.
'''
            ..type = 'GetterReturnModifier',
            member('custom_block')
            ..doc = '''
A single customBlock that will be injected in the public section
of the owning class. For example, if generating code that needs
special getters/setters (e.g. atypical coding pattern) then the
member could be set with *access = ro* and custom accessors may
be provided.
'''
            ..type = 'CodeBlock'
            ..classInit = 'new CodeBlock(null)',
            member('is_streamable')
            ..doc = '''
Indicates member should be streamed if class is streamable.
One of the few flags defaulted to *true*, this flag provides
an opportunity to *not* stream specific members'''
            ..classInit = true,
            member('custom_streamable')
            ..doc = '''
If not-null a custom streamable block. Use this to either hand code or
generate a streamable entry in the containing [Class].
'''
            ..type = 'CodeBlock'
            ..access = RO,
          ],
        ],
        part('class')
        ..enums = [
          enum_('class_code_block')
          ..doc = '''
The various supported code blocks associated with a C++ class. The
name indicates where in the class it appears. Within the class
definition the order is *public*, *protected* then
*private*. Additional locations for just before the class and just
after the class.

So, the following spec:

        (class_('c')
            ..customBlocks = [
              clsPreDecl,
              clsPublic,
              clsProtected,
              clsPrivate,
              clsPostDecl
            ])
        .definition


Gives the following content:

    // custom <ClsPreDecl C>
    // end <ClsPreDecl C>

    class C
    {
    public:
      // custom <ClsPublic C>
      // end <ClsPublic C>

    protected:
      // custom <ClsProtected C>
      // end <ClsProtected C>

    private:
      // custom <ClsPrivate C>
      // end <ClsPrivate C>

    };

    // custom <ClsPostDecl C>
    // end <ClsPostDecl C>
'''
          ..hasLibraryScopedValues = true
          ..values = [
            enumValue(id('cls_open'))
            ..doc = 'The custom block appearing just after class is opened',
            enumValue(id('cls_public'))
            ..doc = 'The custom block appearing in the standard *public* section',
            enumValue(id('cls_protected'))
            ..doc = 'The custom block appearing in the standard *protected* section',
            enumValue(id('cls_private'))
            ..doc = 'The custom block appearing in the standard *private* section',
            enumValue(id('cls_close'))
            ..doc = 'The custom block appearing just before class is closed',
            enumValue(id('cls_pre_decl'))
            ..doc = 'The custom block appearing just before the class definition',
            enumValue(id('cls_post_decl'))
            ..doc = 'The custom block appearing just after the class definition',
          ],
        ]
        ..classes = [
          class_('class_method')
          ..doc = '''
Establishes an interface for generated class methods like
consructors, destructors, overloaded operators, etc.
'''
          ..isAbstract = true
          ..members = [
            member('parent')..type = 'Class'..access = RO,
            member('is_logged')..doc = 'If true add logging'..classInit = false,
            member('template')..type = 'Template'..access = RO,
            member('cpp_access')
            ..doc = 'C++ style access of method'
            ..type = 'CppAccess'
            ..classInit = 'public',
          ],
          class_('default_method')
          ..doc = '''
Unifies the [ClassMethod]s that can be specified as *default*,
like [DefaultCtor], [CopyCtor], etc.

Also provides for *delete*d methods.
'''
          ..isAbstract = true
          ..extend = 'ClassMethod'
          ..members = [
            member('has_custom')
            ..doc = 'Has custom code, so needs protect block'..classInit = false,
            member('top_inject')
            ..doc = '''
Code snippet to inject at beginning of method. The intent is for the
methods to have standard generated implementations, but to also
support programatic injection of implmementation into the
methods. This supports injection near the top of the method.'''
            ..classInit = '',
            member('bottom_inject')
            ..doc = '''
Supports injecting code near the bottom of the method. *See*
*topInject*'''
            ..classInit = '',
            member('uses_default')..classInit = false,
            member('has_delete')..classInit = false,
          ],
          class_('default_ctor')
          ..doc = 'Default ctor, autoinitialized on read'..extend = 'DefaultMethod',
          class_('copy_ctor')
          ..doc = 'Copy ctor, autoinitialized on read'..extend = 'DefaultMethod',
          class_('move_ctor')
          ..doc = 'Move ctor, autoinitialized on read'..extend = 'DefaultMethod',
          class_('assign_copy')..extend = 'DefaultMethod',
          class_('assign_move')..extend = 'DefaultMethod',
          class_('dtor')
          ..doc = 'Provides a destructor'
          ..extend = 'DefaultMethod'
          ..members = [
            member('is_abstract')..classInit = false
          ],
          class_('member_ctor_parm')
          ..doc = '''
A *Member Constructor Parameter*. Defines a single parameter to be passed to a
MemberCtor in order to initialize a single member variable. MemberCtor will
convert strings automatically into instances of MemberCtorParm. For example:

    memberCtor(['x', 'y'])

would produce the following constructor:

    class Point {
    public:
      Point(int x, int y) : x_{x}, y_{y} {}
    private:
      int x_;
      int y_;
    }

But to modify the parameter definition or initializtaion you might rather
construct and modify the MemberCtorParm instead of taking the default:

        final cls = class_('point')
          ..members = [ member('x')..init = 0, member('y')..init = 0 ]
          ..memberCtors = [
            memberCtor( [
              memberCtorParm('x')
              ..defaultValue = '42',
              memberCtorParm('y')
              ..defaultValue = '42'] )
          ];

which produces:

    class Point
    {
    public:
      Point(
        int x = 42,
        int y = 42) :
        x_ { x },
        y_ { y } {
      }

    private:
      int x_ { 0 };
      int y_ { 0 };
    };

'''
          ..hasCtorSansNew = true
          ..members = [
            member('name')
            ..doc = 'Name of member initialized by argument to member ctor'
            ..isFinal = true
            ..ctors = [''],
            member('member')
            ..doc = 'cpp member to be initialized'
            ..type = 'Member',
            member('parm_decl')..doc = '''
*Override* for arguemnt declaration. This is rarely needed. Suppose
you want to initialize member *Y y* from an input argument *X x* that
requires a special function *f* to do the conversion:

    Class(X x) : y_ {f(x)}

Which would be achieved by:

    memberCtor([
      memberCtorParm("y")
      ..parmDecl = "X x"
      ..init = "f(x)"
    ])

''',
            member('init')..doc = '''
*Override* of initialization text. This is rarely needed since
initialization of members in a member ctor is straightforward:

This definition:

    memberCtor(['x', 'y'])

would produce the following constructor:

    class Point {
    public:
      Point(int x, int y) : x_{x}, y_{y} {}
    private:
      int x_;
      int y_;
    }

But sometimes you need more:

    class Umask_scoped_set {
    public:
      Umask_scoped_set(mode_t new_mode) : previous_mode_{umask(new_mode)}
    ...
    }

Which would be achieved by:

    memberCtor([
      memberCtorParm("previous_mode")
      ..parmDecl = "mode_t new_mode"
      ..init = "umask(new_mode)"
    ])

'''
            ..access = WO,
            member('default_value')
            ..doc = '''
If set provides a default value for the parm in the ctor. For example:

    memberCtorParm('x')..defaultValue = '42'

where the type of member *x* is *int* might yield:

    Cls(int x = 42) : x_{x}

''',
          ],
          class_('member_ctor')
          ..doc = '''
Specificication for a member constructor. A member constructor is a constructor
with the intent of initializing one or more members of a class.

Assumig a class has members *int x* and *int y*
*memberCtor(["x", "y"])*

would generate the corresponding:

    Class(int x, int y) : x_{x}, y_{y} {}

If custom logic is additionally required, set the *hasCustom* flag to include a
custom block. In that case the class might look like:

    Class(int x, int y) : x_{x}, y_{y} {
      // custom <Class>
      // end <Class>
    }
'''
          ..extend = 'ClassMethod'
          ..members = [
            member('member_parms')
            ..doc = 'List of members that are passed as arguments for initialization'
            ..type = 'List<MemberCtorParm>'
            ..classInit = [],
            member('decls')
            ..doc = 'List of additional decls ["Type Argname", ...]'
            ..type = 'List<String>',
            member('has_custom')
            ..doc = 'Has custom code, so needs protect block'..classInit = false..access = WO,
            member('custom_label')
            ..doc = 'Label for custom protect block if desired'..access = WO,
            member('has_all_members')
            ..doc = 'If set automatically includes all members as args'
            ..classInit = false,
          ],
          class_('op_equal')
          ..doc = 'Provides *operator==()*'
          ..extend = 'ClassMethod',
          class_('op_less')
          ..doc = 'Provides *operator<()*'
          ..extend = 'ClassMethod',
          class_('op_out')
          ..doc = 'Provides *operator<<()*'
          ..extend = 'ClassMethod'
          ..members = [
            member('uses_indent')
            ..doc = '''
If true uses tls indentation tracking to indent nested
components when streaming'''
            ..classInit = false
          ],
          class_('class')
          ..doc = '''
A C++ class.

Classes optionally have these items:

* A [template]
* A collection of [bases]
* A collection of [members]
* A collection of class local [forwardDecls]
* A collection of class local [usings]
* A collection of class local [enums]
* A collection of class local [forward_ptrs] which are like [usings] but standardized for pointer type
* A collection of *optionally included* standard methods.
  In general these methods are not included unless requested. There
  are two approaches to *requesting* their presence:

  1 - just mention their name (i.e. invoke the getter for the member on the
  class which autoinitializes the member) and the default function will be
  included. This is a *funky* use of function side-effects, but the effect is
  fewer calls required to declaratively describe your class.

  2 - when more configuration of the method is required, call the *with...()*
  function to get scoped access to the function object.

  Example - Case 1:

      print(clangFormat((class_('x')).definition));
      print(clangFormat((class_('x')..copyCtor).definition));
      print(clangFormat((class_('x')..copyCtor.usesDefault = true).definition));

  Prints:

        class X {};

        class X {
         public:
          X(X const& other) {}
        };

        class X {
         public:
          X(X const& other) = default;
        };


  Note that simply naming the copy constructor member of the class will inlude
  its definition. Sometimes you might want to do more with a [ClassMethod]
  definition declaratively in place which is why the *with...()* methods exist.

  Example - Case 2:

      print(clangFormat((
                  class_('x')
                  ..withCopyCtor((ctor) =>
                      ctor..cppAccess = protected
                      /// ... do more with ctor, like inject logging code
                      ))
              .definition));

  Prints:

        class X {
         protected:
          X(X const& other) {}
        };

  The functions are:

  * Optionally included constructors including:

    * [CopyCtor]
    * [MoveCtor]
    * [DefaultCtor]
    * Zero or more member initializing ctors [MemberCtor]

  * Assignment functions:

    * [AssignCopy]
    * [AssignMove]

  * [Dtor]

  * Standard Utility Methods

    * [OpEqual]
    * [OpLess]
    * [OpOut] - Support for streaming fields

* A fixed collection of indexed [codeBlocks] that can be used for
  providing *CustomBlocks* and/or for dynamically injecting code - see
  [CodeBlock].
'''
          ..extend = 'Entity'
          ..mixins = [ 'Testable' ]
          ..members = [
            member('definition')
            ..doc = '''
The contents of the class definition. *Inaccessible* and established
as a member so custom *definition* getter can be called multiple times
on the same class and results lazy-inited here'''
            ..access = IA,
            member('is_struct')
            ..doc = 'Is this definition a *struct*'
            ..classInit = false,
            member('template')
            ..doc = 'The template by which the class is parameterized'
            ..type = 'Template'..access = RO,
            member('forward_decls')
            ..doc = 'List of forward declarations that will appear near the top of the file'
            ..type = 'List<ForwardDecl>'..classInit = [],
            member('const_exprs')
            ..doc = '*constexpr*s associated with the class'
            ..type = 'List<ConstExpr>'..classInit = [],
            member('usings')
            ..doc = '''
List of usings that will be scoped to this class near the top of
the class definition.'''
            ..type = 'List<Using>'..access = RO..classInit = [],
            member('usings_post_decl')
            ..doc = '''
List of usings to occur after the class declaration. Sometimes it is
useful to establish some type definitions directly following the class
so they may be reused among any client of the class. For instance if
class *Foo* will most commonly be used in vector, the using occuring
just after the class definition will work:

    using Foo = std::vector<Foo>;

'''
            ..type = 'List<Using>'..classInit = []..access = RO,
            member('bases')
            ..doc = '''
Base classes this class derives form.
'''
            ..type = 'List<Base>'..classInit = [],
            member('default_ctor')
            ..doc = 'The default constructor'
            ..type = 'DefaultCtor'..access = IA,
            member('copy_ctor')
            ..doc = 'The copy constructor'
            ..type = 'CopyCtor'..access = IA,
            member('move_ctor')
            ..doc = 'The move constructor'
            ..type = 'MoveCtor'..access = IA,
            member('assign_copy')
            ..doc = 'The assignment operator'
            ..type = 'AssignCopy'..access = IA,
            member('assign_move')
            ..doc = 'The assignment move operator'
            ..type = 'AssignMove'..access = IA,
            member('dtor')
            ..doc = 'The destructor'
            ..type = 'Dtor'..access = IA,
            member('member_ctors')
            ..doc = 'A list of member constructors'
            ..type = 'List<MemberCtor>'..classInit = [],
            member('op_equal')..type = 'OpEqual'..access = IA,
            member('op_less')..type = 'OpLess'..access = IA,
            member('op_out')..type = 'OpOut'..access = IA,
            member('forward_ptrs')..type = 'List<PtrType>'..classInit = [],
            member('enums_forward')..type = 'List<Enum>'..classInit = [],
            member('enums')..type = 'List<Enum>'..classInit = [],
            member('members')..type = 'List<Member>'..classInit = [],
            member('friend_class_decls')..type = 'List<FriendClassDecl>'..classInit = [],
            member('custom_blocks')..type = 'List<ClassCodeBlock>'..classInit = [],
            member('is_singleton')..classInit = false,
            member('code_blocks')
            ..access = RO
            ..type = 'Map<ClassCodeBlock, CodeBlock>'..classInit = {},
            member('uses_streamers')
            ..doc = '''
If true adds {using fcs::utils::streamers::operator<<} to streamer.
Also, when set assumes streaming required and [isStreamable]
is *set* as well. So not required to set both.
'''
            ..access = RO
            ..classInit = false,
            member('is_final')
            ..doc = 'If true adds final keyword to class'
            ..classInit = false,
            member('is_immutable')
            ..doc = '''
If true makes all members const provides single member ctor
initializing all.

There are a few options to achieve *immutable* support. The first is
this type, where all fields are constant and therefore must be
initialized. An alternative concept is immutable from perspective of
user. This can be achieved with use of [addFullMemberCtor] and the
developer ensuring the members are not modified. This provides a
stronger guarantee of immutability.'''
            ..classInit = false,
            member('serializers')
            ..doc = 'List of processors supporting flavors of serialization'
            ..type = 'List<Serializer>'
            ..classInit = [],
            member('interface_implementations')
            ..doc = ''
            ..type = 'List<InterfaceImplementation>'
            ..access = RO
            ..classInit = [],
            member('methods')
            ..doc = '''
The [Method]s that are implemented by this [Class]. A [Class]
implements the union of methods in its
[interfaceimplementations]. Each [Method] is identified by its
qualified id *string* which is:

   interface.method_name.signature

Lookup is done by pattern match.
'''
            ..type = 'Map<String, Method>'
            ..access = IA
            ..classInit = {},
            member('cpp_access')
            ..doc = 'A [CppAccess] specifier - only pertinent if class is nested'
            ..type = 'CppAccess'..classInit = 'public',
            member('nested_classes')
            ..doc = 'Classes nested within this class'
            ..type = 'List<Class>'
            ..classInit = [],
            member('pack_align')
            ..doc = r'''
If set, will include *#pragma pack(push, $packAlign)* before the class
and *#pragma pack(pop)* after.
'''
            ..type = 'int',
            member('default_member_access')
            ..doc = 'If set and member has no [access] set, this is used'
            ..type = 'Access',
          ],
        ],
        part('method')
        ..classes = [
          class_('parm_decl')
          ..doc = r"""
A parameter declaration.

Method signatures consist of a List of [ParmDecl] and a return type.
[ParmDecl]s may be constructed from declaration text:

      var pd = new ParmDecl.fromDecl('std::vector< std::vector < double > > matrix');
      print('''
    id    => ${pd.id} (${pd.id.runtimeType})
    type  => ${pd.type}
    ''');

prints:

    id    => matrix (Id)
    type  => std::vector< std::vector < double > >

[ParmDecl]s may be constructed with Id, declaratively:

      var pd = new ParmDecl('matrix')
        ..type = 'std::vector< std::vector < double > >';

      print('''
    id    => ${pd.id} (${pd.id.runtimeType})
    type  => ${pd.type}
    ''');

prints:

    id    => matrix (Id)
    type  => std::vector< std::vector < double > >
"""
          ..extend = 'Entity'
          ..members = [
            member('type'),
          ],
          class_('method_decl')
          ..doc = r"""
A method declaration, which consists of a List of [ParmDecl] (i.e. the
parameters) and a [returnType]

[MethodDecl]s may be constructed from declaration text:

      var md = new MethodDecl.fromDecl('Row_list_t find_row(std::string s)');
      print(md);

prints:

    Row_list_t find_row(std::string s) {
    // custom <find_row>
    // end <find_row>

    }

[MethodDecl]s may be constructed with [id] declaratively:

  var md = new MethodDecl('find_row')
    ..parmDecls = [ new ParmDecl.fromDecl('std::string s') ]
    ..returnType = 'Row_list_t';

prints:

Row_list_t find_row(std::string s) {
// custom <find_row>
// end <find_row>

}
"""
          ..extend = 'Entity'
          ..members = [
            member('parm_decls')..type = 'List<ParmDecl>'..classInit = [],
            member('return_type'),
            member('is_const')
            ..doc = 'True if this [MethodDecl] is *const*'
            ..classInit = false
          ],
          class_('method')
          ..doc = '''
A [Method] represents a single class method that will be *owned* by
the class implementing it. A [Method] method is *owned* by a single
class and therefore has an implementation defined in that class. The
[Method] *has a* signature which it refers to via
[MethodDecl]. [Method] will have its own [CodeBlock] for purpose of
allowing custom code and code insertion.

When defining a class, declaratively or otherwise, [Method]s are
created and owned by the [Class] based on the [implementedInterfaces]
specified. To access the [CodeBlock] of a [Method] in a [Class], use
the [getMethod] function.
'''
          ..members = [
            member('method_decl')..type = 'MethodDecl',
            member('code_block')..type = 'CodeBlock',
            //// TODO: figure best way to support final methods
            // member('is_final')
            // ..doc = 'If true adds final keyword to method'
            // ..classInit = false,
          ],
          class_('interface')
          ..doc = """
A collection of methods that as a group are either virtual or not.  A
*virtual* interface expresses a desire to have code generated that
will implement (i.e. derive from) the set of methods virtually. If the
interface is *not* virtual, it is an indication that the implementers
of the interface will provide implementations to be used via static
polymorphism.

      var md = new Interface('alarmist')
        ..doc = 'Methods that cause alarm'
        ..methodDecls = [
          'void shoutsFireInTheater(int volume)',
          'void wontStopWithTheGlobalWarming()',
          new MethodDecl.fromDecl('void growl()')..doc = 'Scare them'
        ];
      print(md);

prints:

    /**
     Methods that cause alarm
    */
    interface Alarmist
      void shouts_fire_in_theater(int volume) {
        // custom <shouts_fire_in_theater>
        // end <shouts_fire_in_theater>
      }
      void wont_stop_with_the_global_warming() {
        // custom <wont_stop_with_the_global_warming>
        // end <wont_stop_with_the_global_warming>
      }
      /**
       Scare them
      */
      void growl() {
        // custom <growl>
        // end <growl>
      }
    }
"""
          ..extend = 'Entity'
          ..members = [
            member('is_virtual')
            ..doc = '''
If true interface results in pure abstract class, else *static
polymorphic* base.
'''
            ..classInit = false,
            member('method_decls')
            ..type = 'List<MethodDecl>'
            ..access = RO
            ..classInit = [],
          ],
          class_('interface_implementation')
          ..doc = 'An [interface] with a [CppAccess] to be implemented by a [Class]'
          ..members = [
            member('interface')..type = 'Interface',
            member('cpp_access')..type = 'CppAccess'..classInit = 'public',
            member('is_virtual')..classInit = false
          ],
        ],
        part('serializer')
        ..enums = [
          enum_('serialization_style')
          ..doc = 'Serialization using *cereal* supports these types of serialization'
          ..hasLibraryScopedValues = true
          ..values = [
            id('json_serialization'),
            id('xml_serialization'),
            id('binary_serialization'),
          ]
        ]
        ..classes = [
          class_('serializer')
          ..doc = 'Establishes an interface for instance serialization'
          ..isAbstract = true,
          class_('dsv_serializer')
          ..doc = 'Provides support for serialization as *delimited separated values*'
          ..implement = [ 'Serializer' ]
          ..members = [
            member('delimiter')..classInit = ':',
          ],
          class_('cereal')
          ..doc = 'Adds support for serialization using *cereal*'
          ..implement = [ 'Serializer' ]
          ..members = [
            member('styles')..type = 'List<SerializationStyle>'..classInit = [],
          ],
        ],
        part('header')
        ..classes = [
          class_('header')
          ..doc = 'A single c++ header'
          ..extend = 'CppFile'
          ..members = [
            member('is_api_header')
            ..doc = '''
If true marks this header as special to the set of headers in its library in that:
(1) It will be automatically included by all other headers
(2) For windows systems it will be the place to provide the api decl support
(3) Will have code that initializes the api
'''
            ..classInit = false,
          ],
        ],
        part('impl')
        ..classes = [
          class_('impl')
          ..doc = 'A single implementation file (i.e. *cpp* file)'
          ..extend = 'CppFile'
          ..members = [
          ]
        ],
        part('test_provider')
        ..enums = [
          enum_('tc_code_block')
          ..doc = '''
The various supported code blocks associated with *TestClause*.
The *TestClauses* are modeled after the *Catch* library *BDD* approach.
'''
          ..hasLibraryScopedValues = true
          ..values = [
            enumValue(id('tc_open'))
            ..doc = 'The custom block appearing at the start of the clause',
            enumValue(id('tc_close'))
            ..doc = 'The custom block appearing at the end of the clause',
          ],
        ]
        ..classes = [
          class_('test_clause')
          ..isAbstract = true
          ..extend = 'Entity'
          ..doc = '''
Models common elements of the *Given*, *When*, *Then* clauses.
Each *TestClause* has its own [clause] text associated with it
and [CodeBlock]s to augment/initialize/teardown.
'''
          ..members = [
            member('start_code_block')
            ..type = 'CodeBlock'
            ..classInit = 'new CodeBlock(null)',
            member('end_code_block')
            ..type = 'CodeBlock'
            ..classInit = 'new CodeBlock(null)',
          ],
          class_('then')
          ..extend = 'TestClause'
          ..members = [
            member('is_and')
            ..classInit = false
          ],
          class_('when')
          ..extend = 'TestClause'
          ..members = [
            member('thens')
            ..type = 'List<Then>'
          ],
          class_('given')
          ..extend = 'TestClause'
          ..members = [
            member('whens')
            ..type = 'List<When>'
          ],
          class_('test_scenario')
          ..extend = 'Entity'
          ..members = [
            member('givens')..type = 'List<Given>',
          ],

          class_('testable')
          ..members = [
            member('test_scenarios')..type = 'List<TestScenario>'..classInit = [],

            member('test')
            ..doc = 'The single test for this [Testable]'
            ..type = 'Test'
            ..access = WO,

            member('includes_test')
            ..doc = 'If set will provide a blank test for the [Testable]'
            ..classInit = false,

            member('test_functions')
            ..doc = 'Strings naming functions to be included in the test'
            ..access = WO
            ..type = 'List<String>',

          ],

          class_('test_provider')
          ..isAbstract = true
          ..members = [],

          class_('boost_test_provider')
          ..extend = 'TestProvider',

          class_('catch_test_provider')
          ..extend = 'TestProvider'
          ..members = [
          ],
        ],
        part('lib')
        ..enums = [
          enum_('file_code_block')
          ..doc = '''
Set of pre-canned blocks where custom or generated code can be placed.
The various supported code blocks associated with a C++ file. The
name indicates where in the file it appears.

So, the following spec:

    final h = header('foo')
      ..includes = [ 'iostream' ]
      ..namespace = namespace(['foo'])
      ..customBlocks = [
        fcbCustomIncludes,
        fcbPreNamespace,
        fcbBeginNamespace,
        fcbEndNamespace,
        fcbPostNamespace,
      ];
    print(h.contents);

prints:

    #ifndef __FOO_FOO_HPP__
    #define __FOO_FOO_HPP__

    #include <iostream>

    // custom <FcbCustomIncludes foo>
    // end <FcbCustomIncludes foo>

    // custom <FcbPreNamespace foo>
    // end <FcbPreNamespace foo>

    namespace foo {
      // custom <FcbBeginNamespace foo>
      // end <FcbBeginNamespace foo>

      // custom <FcbEndNamespace foo>
      // end <FcbEndNamespace foo>

    } // namespace foo
    // custom <FcbPostNamespace foo>
    // end <FcbPostNamespace foo>

    #endif // __FOO_FOO_HPP__

'''
          ..hasLibraryScopedValues = true
          ..values = [
            enumValue(id('fcb_pre_includes'))
            ..doc = '''
Custom block any code just before includes begin
Useful for putting definitions just prior to includes, e.g.

    #define CATCH_CONFIG_MAIN
    #include "catch.hpp"
''',
            enumValue(id('fcb_custom_includes'))
            ..doc = 'Custom block for any additional includes appearing just after generated includes',
            enumValue(id('fcb_pre_namespace'))
            ..doc = 'Custom block appearing just before the namespace declaration in the code',
            enumValue(id('fcb_begin_namespace'))
            ..doc = 'Custom block appearing at the begining of and inside the namespace',
            enumValue(id('fcb_end_namespace'))
            ..doc = 'Custom block appearing at the end of and inside the namespace',
            enumValue(id('fcb_post_namespace'))
            ..doc = 'Custom block appearing just after the namespace declaration in the code',
          ]
        ]
        ..classes = [
          class_('lib')
          ..doc = 'A c++ library'
          ..extend = 'Entity'
          ..mixins = [ 'Testable' ]
          ..implement = [ 'CodeGenerator' ]
          ..members = [
            member('namespace')..type = 'Namespace'..classInit = 'new Namespace()',
            member('headers')..type = 'List<Header>'..classInit = [],
            member('impls')..type = 'List<Impl>'..classInit = [],
            member('tests')..type = 'List<Test>'..classInit = [],
          ],
        ],
        part('app')
        ..enums = [
          enum_('arg_type')
          ..doc = '''
Set of argument types supported by command line option processing.
'''
          ..requiresClass = true
          ..values = [
            enumValue(id('int'))
            ..doc = 'The command line arg is an integer',
            enumValue(id('double'))
            ..doc = 'The command line arg is a double',
            enumValue(id('string'))
            ..doc = 'The command line arg is a string',
            enumValue(id('flag'))
            ..doc = 'The command line arg is a flag - i.e. a boolean',
          ]
        ]
        ..classes = [
          class_('app_arg')
          ..doc = '''
Metadata associated with an argument to an application.  Requires and
geared to features supported by boost::program_options. The supporting
code for arguments is spread over a few places in the main file of an
[App]. Examples of declarations follow:

      print(br([
        appArg('filename')
        ..shortName = 'f',

        appArg('in_file')
        ..shortName = 'f'
        ..defaultValue = 'input.txt',

        appArg('pi')
        ..shortName = 'p'
        ..isRequired = true
        ..defaultValue = 3.14,

        appArg('source_file')
        ..shortName = 's'
        ..isMultiple = true
      ]));

Prints:

    AppArg(filename)
      argType: String
      cppType: std::string
      flagDecl: "filename,f"
      isRequired: false
      isMultiple: false
      defaultValue: null

    AppArg(in_file)
      argType: String
      cppType: std::string
      flagDecl: "in-file,f"
      isRequired: false
      isMultiple: false
      defaultValue: input.txt

    AppArg(pi)
      argType: Double
      cppType: double
      flagDecl: "pi,p"
      isRequired: true
      isMultiple: false
      defaultValue: 3.14

    AppArg(source_file)
      argType: String
      cppType: std::vector< std::string >
      flagDecl: "source-file,s"
      isRequired: false
      isMultiple: true
      defaultValue: null


For an [App], if no [Arg] in [args] is named *help* or has [shortName]
of *h* then the following *help* argument is provided. The help text
will include the doc string of the [App].

      args.insert(0, new AppArg(new Id('help'))
        ..shortName = 'h'
        ..defaultValue = false
        ..descr = 'Display help information');


Example: Here are the [AppArg]s of an simple application:

    arg('timestamp')
    ..shortName = 't'
    ..descr = 'Some form of timestamp'
    ..isMultiple = true
    ..type = ArgType.STRING,
    arg('date')
    ..shortName = 'd'
    ..descr = 'Some form of date'
    ..isMultiple = true
    ..type = ArgType.STRING,

When run, the help looks something like:

    App for converting between various forms of date/time

    AllowedOptions:
      -h [ --help ]          Display help information
      -t [ --timestamp ] arg Some form of timestamp
      -d [ --date ] arg      Some form of date
'''
          ..extend = 'Entity'
          ..members = [
            member('type')..type = 'ArgType'..classInit = 'ArgType.STRING',
            member('short_name'),
            member('is_multiple')..classInit = false,
            member('is_required')..classInit = false,
            member('default_value')..type = 'Object'..access = RO,
          ],
          class_('app')
          ..doc = '''

A C++ application. Application related files are generated in location based on
[namespace] the namespace. For example, the following code:

    app('date_time_converter')
      ..namespace = namespace(['fcs'])
      ..args = [
        arg('timestamp')
        ..shortName = 't'
        ..descr = 'Some form of timestamp'
        ..isMultiple = true
        ..type = ArgType.STRING,
        arg('date')
        ..shortName = 'd'
        ..descr = 'Some form of date'
        ..isMultiple = true
        ..type = ArgType.STRING,
      ];

will generate a C++ file containing *main* at location:

    \$root/cpp/app/date_time_converter/date_time_converter.cpp

Since [App] extends [Impl] it supports local instances of
[constExprs] [usings], [enums], [forwardDecls], and [classes],
as well as [headers] and [impls] which may be part of the
application and not necessarily suited for a separate library.
'''
          ..extend = 'Impl'
          ..implement = [ 'CodeGenerator' ]
          ..members = [
            member('args')
            ..doc = 'Command line arguments specific to this application'
            ..type = 'List<AppArg>'..classInit = [],
            member('namespace')
            ..doc = 'Namespace associated with application code'
            ..type = 'Namespace'..access = IA,
            member('headers')
            ..doc = '''
Additional headers that are associated with the application itself, as
opposed to belonging to a reusable library.'''
            ..type = 'List<Header>'..classInit = [],
            member('impls')
            ..doc = '''
Additional implementation files associated with the
application itself, as opposed to belonging to a reusable
library.'''
            ..type = 'List<Impl>'..classInit = [],
            member('required_libs')
            ..doc = '''
Libraries required to build this executable. *Warning* potentially
deprecated in the future. Originally when generating boost jam files
it was convenient to associate the required libraries directly in the
code generation scripts. With cmake it was simpler to just incorporate
protect blocks where the required libs could be easily added.
'''
            ..type = 'List<String>'..classInit = [],
            member('builders')
            ..doc = '''
List of builders to generate build scripts of a desired flavor (cmake
is only one supported at this time)'''
            ..type = 'List<AppBuilder>'..classInit = [],
            member('main_code_block')
            ..doc = '''
An App is an Impl and therefore contains accesors to FileCodeBlock
sections (e.g. fcbBeginNamespace, fcbPostNamespace, ...). The heart of
an application impl file is the main, so this [CodeBlock] supports
injecting code in main
'''
            ..type = 'CodeBlock'
            ..classInit = "new CodeBlock('main')",
          ],
          class_('app_builder')
          ..doc = '''
Base class establishing interface for generating build scripts for
libraries, apps, and tests'''
          ..isAbstract = true
          ..implement = [ 'CodeGenerator' ]
          ..members = [
            member('app')..type = 'App'
          ],
        ],
        part('cmake_support')
        ..classes = [
          class_('cmake_installation_builder')
          ..extend = 'InstallationBuilder',
        ],
        part('script')
        ..classes = [
          class_('script')
          ..extend = 'Entity'
          ..implement = [ 'CodeGenerator' ]
          ..members = [
          ],
        ],
        part('test')
        ..classes = [
          class_('test')
          ..extend = 'Impl'
          ..implement = [ 'CodeGenerator' ]
          ..members = [
            member('testable')..type = 'Testable',
            member('headers')..type = 'List<Header>'..classInit = [],
            member('impls')..type = 'List<Impl>'..classInit = [],
            member('test_implementations')..type = 'Map<String, String>'..classInit = {}..access = RO,
            member('required_libs')..type = 'List<String>'..classInit = [],
          ],
        ],
        part('installation')
        ..classes = [
          class_('installation_builder')
          ..isAbstract = true
          ..implement = [ 'CodeGenerator' ]
          ..doc = '''
Creates builder for an installation (ie ties together all build artifacts)
'''
          ..members = [
            member('installation')..type = 'Installation'
          ],
          class_('installation')
          ..extend = 'Entity'
          ..implement = [ 'CodeGenerator' ]
          ..members = [
            member('root')..doc = 'Fully qualified path to installation'..access = RO,
            member('paths')..type = 'Map<String, String>'..classInit = {}..access = RO,
            member('libs')..type = 'List<Lib>'..classInit = [],
            member('apps')..type = 'List<App>'..classInit = [],
            member('tests')..type = 'List<Test>'..classInit = [],
            member('scripts')..type = 'List<Script>'..classInit = [],
            member('test_provider')
            ..doc = 'Provider for generating tests'
            ..type = 'TestProvider'
            ..classInit = 'new CatchTestProvider()',
            member('builders')
            ..doc = 'List of builders for the installation (cmake is only one supported at this time)'
            ..type = 'List<InstallationBuilder>'..classInit = [],
            member('includes_header_compiles')
            ..doc = '''
If set will generate a single cpp per header that includes that header.
If that cpp compiles it is an indication the header is doing a proper
job of including all its dependencies. If there are compile errors,
revisit the header and add required includes'''
            ..classInit = false,
            member('doxy_config')
            ..type = 'DoxyConfig'
            ..classInit = 'new DoxyConfig()',
          ],
          class_('path_locator')
          ..members = [
            member('env_var')
            ..doc = 'Environment variable specifying location of path, if set this path is used'
            ..isFinal = true,
            member('default_path')
            ..doc = 'Default path for the item in question'
            ..isFinal = true,
            member('path')..access = RO,
          ],
        ],

        part('doxy')
        ..classes = [
          class_('doxy_config')
          ..members = [
            member('project_name')..classInit = '',
            member('project_brief')..classInit = '',
            member('output_directory')..classInit = '',
            member('input')..classInit = '',
            member('exclude')..classInit = '',
            member('example_path')..classInit = '',
            member('image_path')..classInit = '',
            member('html_stylesheet')..classInit = '',
            member('chm_file')..classInit = '',
            member('include_path')..classInit = '',
          ]
        ]
      ],

      library('hdf5_support')
      ..imports = [
        'package:ebisu_cpp/ebisu_cpp.dart',
        'package:id/id.dart',
      ]
      ..doc = 'Provide C++ classes support for reading/writing to hdf5 packet table'
      ..parts = [
        part('packet_table')
        ..classes = [
          class_('class_not_found_exception')
          ..doc = '''
Indicates a class could not be found in the [Installation] for adding
hdf5 packet table support
'''
          ..implement = ['Exception']
          ..members = [
            member('message')
            ..isFinal = true
            ..access = RO
            ..doc = 'Exception details',
          ],
          class_('log_group')
          ..hasCtorSansNew = true
          ..members = [
            member('class_name')
            ..doc = 'Name of class, *snake case*, to add a packet table log group'
            ..ctors = ['']
            ..isFinal = true,
            member('member_names')
            ..doc = '''
Name of members of class, *snake case*, to include in the packet table
log group. An empty list will include all members in the table.
'''
            ..type = 'List<String>'
            ..ctorInit = 'const []'
            ..ctorsOpt = ['']
            ..isFinal = true,
          ],
          class_('packet_table_decorator')
          ..isImmutable = true
          ..hasCtorSansNew = true
          ..implement = [
            'InstallationDecorator',
          ]
          ..members = [
            member('log_groups')..type = 'List<LogGroup>'
          ]
        ]
      ],

    ];

  ebisu.scripts = [
    script('bootstrap_ebisu_cpp')
    ..imports = [
      'package:id/id.dart',
      'package:path/path.dart',
      "'package:ebisu/ebisu.dart' as ebisu",
      'package:ebisu/ebisu_dart_meta.dart',
    ]
    ..doc = 'Creates an ebisu_cpp setup'
    ..classes = [
      class_('project')
      ..hasJsonToString = true
      ..members = [
        member('id')..type = 'Id',
        member('root_path'),
        member('codegen_path'),
        member('script_name'),
        member('ebisu_file_path'),
        member('cpp_file_path'),
      ]
    ]
    ..args = [
      scriptArg('project_path')
      ..doc = 'Path to top level of desired ebisu project'
      ..abbr = 'p',
      scriptArg('add_app')
      ..doc = 'Add library to project'
      ..abbr = 'a',
      scriptArg('add_lib')
      ..doc = 'Add library to project'
      ..abbr = 'l',
      scriptArg('add_script')
      ..doc = 'Add script to project'
      ..abbr = 's',
    ]
  ];

  ebisu.generate();
}

////////////////////////////////////////////////////////////////////////////////
// Large doc comment for the cpp library
final cppLibraryDoc = '''
Library to facilate generation of c++ code.

The intent is to get as declarative as possible with the specification of C++
entities to make code generation as simple and fun as possible. The primary
focus of these utilities is in generating the *structure* of c++ code. This is
achieved by modeling the C++ language at a relatively high level and selectively
choosing what parts of the language lend themselves to the approach.

For sample code that uses this library to generate its structure see:
[fcs project](https://github.com/patefacio/fcs)

For a small taste, the following is the current description of a small C++
library called *raii* which provides a few utilities for handling the *resource
acquisition is initialization* idiom.

    import 'package:ebisu_cpp/cpp.dart';
    import '../../lib/installation.dart';

    final raii = lib('raii')
      ..namespace = namespace([ 'fcs', 'raii' ])
      ..headers = [
        header('change_tracker')
        ..includes = [ 'boost/call_traits.hpp' ]
        ..classes = [
          class_('change_tracker')
          ..descr = \'\'\'
    Tracks current/previous values of the given type of data. For some
    algorithms it is useful to be able to examine/perform logic on
    current value and compare or evalutate how it has changed since
    previous value.\'\'\'
          ..template = [ 'typename T' ]
          ..customBlocks = [clsPublic]
          ..members = [
            member('current')..type = 'T'..access = ro,
            member('previous')..type = 'T'..access = ro,
          ],
          ...
        ],
        header('api_initializer')
        ..test.customBlocks = [ fcbPreNamespace ]
        ..test.includes.addAll(['vector', 'fcs/utils/streamers/containers.hpp', ])
        ..includes = [ 'list', 'map', 'memory' ]
        ..usings = [
          'Void_func_t = void (*)(void)',
        ]
        ..classes = [
          class_('functor_scope_exit')
          ..includesTest = true
          ..template = [ 'typename FUNCTOR = Void_func_t' ]
          ..usings = [ 'Functor_t = FUNCTOR' ]
          ..customBlocks = [ clsPublic ]
          ..memberCtors = [ memberCtor(['functor']) ]
          ..members = [
            member('functor')..type = 'Functor_t'..hasNoInit = true..access = ro,
          ],
          ...
          class_('api_initializer')
          ..usings = [
            'Api_initializer_registry_t = Api_initializer_registry< INIT_FUNC, UNINIT_FUNC >'
          ]
          ..template = [
            'typename INIT_FUNC = Void_func_t',
            'typename UNINIT_FUNC = Void_func_t',
          ]
          ..customBlocks = [ clsPublic ]
        ]
      ];

    addItems() => installation.addLib(raii);

    main() {
      addItems();
      installation.generate();
    }

When that script is run, the following is output:

    No change: \$TOP/fcs/cpp/fcs/raii/change_tracker.hpp
    No change: \$TOP/fcs/cpp/fcs/raii/api_initializer.hpp
    No change: \$TOP/fcs/cpp/tests/fcs/raii/test_change_tracker.cpp
    No change: \$TOP/fcs/cpp/tests/fcs/raii/test_api_initializer.cpp

So when the script is run the code is *regenerated* and any changed files will
be indicated as such. In this case, since the code was previously generated, it
indicates there were no updates.
''';
